/*
 *
 * Copyright (c) 2019 Texas Instruments Incorporated
 *
 * All rights reserved not granted herein.
 *
 * Limited License.
 *
 * Texas Instruments Incorporated grants a world-wide, royalty-free, non-exclusive
 * license under copyrights and patents it now or hereafter owns or controls to make,
 * have made, use, import, offer to sell and sell ("Utilize") this software subject to the
 * terms herein.  With respect to the foregoing patent license, such license is granted
 * solely to the extent that any such patent is necessary to Utilize the software alone.
 * The patent license shall not apply to any combinations which include this software,
 * other than combinations with devices manufactured by or for TI ("TI Devices").
 * No hardware patent is licensed hereunder.
 *
 * Redistributions must preserve existing copyright notices and reproduce this license
 * (including the above copyright notice and the disclaimer and (if applicable) source
 * code license limitations below) in the documentation and/or other materials provided
 * with the distribution
 *
 * Redistribution and use in binary form, without modification, are permitted provided
 * that the following conditions are met:
 *
 * *       No reverse engineering, decompilation, or disassembly of this software is
 * permitted with respect to any software provided in binary form.
 *
 * *       any redistribution and use are licensed by TI for use only with TI Devices.
 *
 * *       Nothing shall obligate TI to provide you with source code for the software
 * licensed and provided to you in object code.
 *
 * If software source code is provided to you, modification and redistribution of the
 * source code are permitted provided that the following conditions are met:
 *
 * *       any redistribution and use of the source code, including any resulting derivative
 * works, are licensed by TI for use only with TI Devices.
 *
 * *       any redistribution and use of any object code compiled from the source code
 * and any resulting derivative works, are licensed by TI for use only with TI Devices.
 *
 * Neither the name of Texas Instruments Incorporated nor the names of its suppliers
 *
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * DISCLAIMER.
 *
 * THIS SOFTWARE IS PROVIDED BY TI AND TI'S LICENSORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TI AND TI'S LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef TIVX_LIDAR_NODES_H_
#define TIVX_LIDAR_NODES_H_

#include <VX/vx.h>

#ifdef __cplusplus
extern "C" {
#endif

/*! \brief Creates a Lidar Ground Plane Classification (GPC) Node.
 *
 * The node receives a 3d point cloud, estimates normal vectors and determines which points
 * are believed to be on the ground plane based on their normal vector. Points are tagged
 * as "on the ground" or "not on the ground" in the output 3d point cloud.
 *
 * In addition, the output point cloud is transformed to the current INS position with the origin
 * provided by root_ecef.
 *
 * Therefore, PTK_INS module needs to be enabled in the system.
 *
 * \param [in] graph             The reference to the graph.
 * \param [in] configuration     Kernel configuration. Array element type tivx_lidar_gpc_config_t
 * \param [in] point_cloud_in    Array element type PTK_PointCloud
 * \param [in] lidar_meta        Array element type PTK_LidarMeta
 * \param [in] m_ego_lidar       Array element type PTK_RigidTransform
 * \param [in] root_ecef         Array element type PTK_Position
 * \param [out] point_cloud_out  Array element type PTK_PointCloud
 * \param [out] normal_cloud     Array element type PTK_PointCloud
 *
 * \ingroup group_vision_apps_kernels_lidar
 *
 * \retval vx_node A node reference. Any possible errors preventing a successful creation should be checked using <tt>vxGetStatus</tt>
 */
VX_API_ENTRY vx_node VX_API_CALL tivxLidarGpcNode(vx_graph graph,
                                      vx_user_data_object  configuration,
                                      vx_user_data_object  point_cloud_in,
                                      vx_user_data_object  lidar_meta,
                                      vx_user_data_object  m_ego_lidar,
                                      vx_user_data_object  root_ecef,
                                      vx_user_data_object  point_cloud_out,
                                      vx_user_data_object  normal_cloud);

/*! \brief Creates a Lidar Motion Distortion Correction (MDC) Node.
 *
 * The node receives a raw lidar 3d point cloud and uses INS motion measurements together with the lidar-to-INS(ego) calibration
 * to correct the 3d point cloud for motion distortion caused by the rotation of the Velodyne lidar.
 *
 * PTK_INS module needs to be enabled in the system.
 *
 * \param [in] graph             The reference to the graph.
 * \param [in] point_cloud       Array element type PTK_PointCloud
 * \param [in] meta              Array element type PTK_LidarMeta
 * \param [in] m_ego_lidar       Array element type PTK_RigidTransform
 * \param [in] root_ecef         Array element type PTK_Position
 * \param [out] mdc_point_cloud  Array element type PTK_PointCloud
 *
 * \ingroup group_vision_apps_kernels_lidar
 *
 * \retval vx_node A node reference. Any possible errors preventing a successful creation should be checked using <tt>vxGetStatus</tt>
 */
VX_API_ENTRY vx_node VX_API_CALL tivxLidarMdcNode(vx_graph graph,
                                      vx_user_data_object  point_cloud,
                                      vx_user_data_object  meta,
                                      vx_user_data_object  m_ego_lidar,
                                      vx_user_data_object  root_ecef,
                                      vx_user_data_object  mdc_point_cloud);

#ifdef __cplusplus
}
#endif

#endif /* TIVX_LIDAR_NODES_H_ */


